UNIT Falltuer;

(*------------------------------------------------------------------------*)
(* Die Falltrfunktion als mathematische Grundlage fr eine Codierung und *)
(* Decodierung auf dem Kleincomputer -- LOG IN 4 (1984) Heft 1 bis Heft 3 *)
(*------------------------------------------------------------------------*)

INTERFACE (*-------------------------------------------------------------*)

CONST
  KeyLength = 16;  (* Schlssel-Lnge *)

TYPE
  CodeT   = LONGINT;
  VectorT = ARRAY[1..KeyLength] OF CodeT;

PROCEDURE AufbauKomponenten(VAR vector : VectorT);
(* Erstellt einen Vektor nach: vector(count+1):= 2 * vector(count) + add *)

PROCEDURE Bestimme(vector : VectorT; VAR divisor, einheit, invers : CodeT);
(* Bestimmt den Divisor, die Einheit und das inverse Element *)

PROCEDURE AufbauOeffentlicherVektor(vector : VectorT; divisor, invers : CodeT; VAR open : VectorT);
(* Liefert aus dem geheimen Vektor, Divisor und inversem Element *)
(* den passenden ffentlichen Vektor                             *)

FUNCTION Verschluesseln(open : VectorT; klarTextFile, codeTextFile : STRING) : BOOLEAN;
(* Liest den Text in klarTextFile, verschlsselt ihn ber den ffentlichen *)
(* Vektor open und schreibt den codierten Text in codeTextFile.            *)
(* Liefert TRUE, wenn alles geklappt hat, sonst FALSE.                     *)

FUNCTION Entschluesseln(vector : VectorT; divisor, einheit : CodeT;
                         codeTextFile, klarTextFile : STRING) : BOOLEAN;
(* Liest den Code in codeTextFile, entschlsselt ihn ber den geheimen *)
(* Vektor vector und schreibt den Klartext in klarTextFile.            *)
(* Liefert TRUE, wenn alles geklappt hat, sonst FALSE.                 *)

IMPLEMENTATION (*--------------------------------------------------------*)

PROCEDURE AufbauKomponenten(VAR vector : VectorT);
(* Erstellt einen Vektor nach: vector(count+1):= 2 * vector(count) + add *)
VAR
  add   : CodeT;
  count : INTEGER;

  PROCEDURE GetAdd(number: INTEGER; VAR add: CodeT);
  (* Liest additives Glied ein *)
  BEGIN
    Write('Komponente(',number:2,'): ');
    ReadLn(add);
  END;

BEGIN
  count:= 1;
  REPEAT
    IF ( count <> 1 ) THEN
      vector[count]:= vector[count-1] * 2
    ELSE
      vector[count]:= 1;
    GetAdd(count,add);
    IF ( add < 0 ) THEN
      add:= 1;
    vector[count]:= vector[count] + add;
    INC(count);
  UNTIL ( count > KeyLength )
END;

PROCEDURE Bestimme(vector : VectorT; VAR divisor, einheit, invers : CodeT);
(* Bestimmt den Divisor, die Einheit und das inverse Element *)

VAR
  step : CodeT;  (* Schrittweite *)

  FUNCTION EinheitOkay(einheit : CodeT) : BOOLEAN;
  (* Prft, ob die Einheit grer als Null und teilerfremd zum Divisor ist *)

    FUNCTION Teilerfremd(zahl1, zahl2 : CodeT) : BOOLEAN;
    (* Prft, ob zahl1 und zahl2 teilerfremd sind (Euklid'scher Algorithmus) *)
    BEGIN
      IF ( zahl2 = 0 ) THEN
        IF ( zahl1 = 1 ) THEN
          Teilerfremd:= TRUE
        ELSE
          Teilerfremd:= FALSE
      ELSE
        Teilerfremd:= Teilerfremd(zahl2,zahl1-zahl2*(zahl1 DIV zahl2));
    END;

  BEGIN
    EinheitOkay:= ( (einheit > 0) AND (einheit < divisor) AND teilerfremd(einheit,divisor) );
  END;

BEGIN
  (* Divisor einlesen *)
  REPEAT
    Write('Bitte den Divisor (grer als ',vector[KeyLength],') eingeben: ');
    ReadLn(divisor);
    WriteLn;
  UNTIL ( divisor > vector[KeyLength] );
  REPEAT
    (* Einheit einlesen *)
    REPEAT
      Write('Bitte die Einheit (grer Null und teilerfremd mit dem Divisor) eingeben: ');
      ReadLn(einheit);
    UNTIL EinheitOkay(einheit);
    (* Suche des inversen Elements *)
    WriteLn;
    Write('Suche das inverse Element...');
    WriteLn;
    step:= divisor DIV einheit;
    invers:= 1;
    REPEAT
      INC(invers,step);
      Write(invers,' | ');
    UNTIL ( ( einheit * invers MOD divisor ) = 1 ) OR (invers > divisor );
    WriteLn;
    WriteLn;
  UNTIL ( invers < divisor );
END;

PROCEDURE AufbauOeffentlicherVektor(vector : VectorT; divisor, invers : CodeT; VAR open : VectorT);
(* Liefert aus dem geheimen Vektor, Divisor und inversem Element *)
(* den passenden ffentlichen Vektor                             *)
VAR
  count : INTEGER;
BEGIN
  count:= 1;
  WHILE ( count <= KeyLength ) DO
  BEGIN
    open[count]:= ( vector[count] * invers ) MOD divisor;
    INC(count);
  END;
END;

FUNCTION Power(zahl : REAL; hoch : INTEGER) : REAL;
(* Berechnet die Potenz zahl ^ hoch, wobei hoch >= 0           *)
(* Algorithmus aus: Pascal User Manual von K. Jensen, N. Wirth *)
VAR
  ergebnis : REAL;
BEGIN
  ergebnis:= 1.0;
  WHILE ( hoch > 0 ) DO
  BEGIN
    WHILE NOT ODD(hoch) DO
    BEGIN
      hoch:= hoch DIV 2;
      zahl:= SQR(zahl);
    END;
      hoch:= hoch - 1;
      ergebnis:= zahl * ergebnis;
  END;
  Power:= ergebnis;
END;

FUNCTION Exists(filename : STRING) : BOOLEAN;
(* Wenn die Datei in filename existiert liefert die Funktion TRUE, sonst FALSE *)
VAR
  datei  : FILE;
BEGIN
  {$I-}
  Assign(datei,filename);
  Reset(datei);
  Close(datei);
  {$I+}
  Exists:= (ioresult = 0);
END;

FUNCTION Verschluesseln(open : VectorT; klarTextFile, codeTextFile : STRING) : BOOLEAN;
(* Liest den Text in klarTextFile, verschlsselt ihn ber den ffentlichen *)
(* Vektor open und schreibt den codierten Text in codeTextFile.            *)
(* Liefert TRUE, wenn alles geklappt hat, sonst FALSE.                     *)
VAR
  zeichenCount       : INTEGER;
  code               : CodeT;
  klarFile           : FILE OF CHAR;
  codeFile           : FILE OF CodeT;
  klar               : STRING;
  zeichen            : CHAR;

  PROCEDURE Codieren(klar : STRING; VAR code : CodeT);
  (* Codiert das "Wort" in klar und legt das Ergebnis in code ab *)
  CONST
    WordLength = KeyLength DIV 8; (* Wortlnge bei 8 Bits pro Zeichen *)
  TYPE
    ByteT = ARRAY[1..8] OF BOOLEAN;    (* Zhler hier mal von 1 bis 8 *)
    BitFolgeT = ARRAY[1..KeyLength] OF BOOLEAN;
  VAR
    bitfolge : BitFolgeT;

    PROCEDURE String2Bitfolge(str : STRING; VAR bitfolge : BitFolgeT);
    (* Konvertiert die Zeichenfolge in eine Bitfolge *)
    VAR
      zeichenCount, count, position : INTEGER;
      byte                          : ByteT;

      PROCEDURE Char2Byte(zeichen : CHAR; VAR byte : ByteT);
      (* Macht aus einem Zeichen eine Bitfolge von 8 Bits *)
      VAR
        ord_zahl, count, calc : INTEGER;
      BEGIN
        ord_zahl:= ORD(zeichen);
        FOR count:= 8 DOWNTO 1 DO
        BEGIN
          (* Prfen, ob das Bit zu setzen ist *)
          calc:= ROUND( Power(2.0,(count-1)) );
          IF ( calc <= ord_zahl ) THEN
          BEGIN
            byte[count]:= TRUE;
            DEC(ord_zahl,calc);
          END
          ELSE
            byte[count]:= FALSE;
        END;
      END;

    BEGIN
      (* Fr alle Zeichen in der Zeichenfolge *)
      FOR zeichenCount:= 1 TO WordLength DO
      BEGIN
        (* Wandle das Zeichen in ein "Byte" um *)
        Char2Byte(str[zeichenCount],byte);
        (* Bestimme die Position des Byte dieses Zeichens in der Bitfolge *)
        position:= zeichenCount * 8 - 8;
        (* Fge das Byte in die Bitfolge ein *)
        FOR count:= 1 TO 8 DO
          bitfolge[position+count]:= byte[count];
      END;
    END;

    PROCEDURE Bitfolge2Code(bitfolge : BitFolgeT; VAR code : CodeT);
    (* Erstellt aus gegebener Bitfolge & ffentlichem Vektor den Code *)
    VAR
      bitCount : INTEGER;
    BEGIN
      code:= 0;
      FOR bitCount:= 1 TO KeyLength DO
        IF bitfolge[bitCount] THEN
          code:= code + open[bitCount];
    END;

  BEGIN
    (* Mache Bitfolge aus dem Wort *)
    String2Bitfolge(klar,bitfolge);
    (* Mache Code aus der Bitfolge *)
    Bitfolge2Code(bitfolge,code);
  END;

BEGIN
  (* Prfen, ob Klartext-Datei existiert *)
  IF NOT Exists(klarTextFile) THEN
  BEGIN
    (* Negative Rckmeldung *)
    Verschluesseln:= FALSE;
    Exit;
  END;
  (* Dateien aufmachen *)
  (* FIO.EOF:= FALSE; *)
  Assign(klarFile,klarTextFile);
  Reset(klarFile);
  Assign(codeFile,codeTextFile);
  Rewrite(codeFile);
  (* Codierungsschleife *)
  WHILE NOT EOF(klarFile) DO
  BEGIN
    (* "Wort" lesen (klar) *)
    klar:= '';
    FOR zeichenCount:= 1 TO (KeyLength DIV 8) DO
    BEGIN
      IF NOT EOF(klarFile) THEN
      BEGIN
        Read(klarFile,zeichen);
        klar:= Concat(klar,zeichen);
      END;
    END;
    (* Codieren *)
    Write('Codiere ',klar,' --> ');
    Codieren(klar,code);
    WriteLn(code);
    (* Code rausschreiben *)
    Write(codeFile,code);
  END;
  (* Dateien schlieen *)
  Close(klarFile);
  Close(codeFile);
  (* Positive Rckmeldung *)
  Verschluesseln:= TRUE;
END;

FUNCTION Entschluesseln(vector : VectorT; divisor, einheit : CodeT;
                         codeTextFile, klarTextFile : STRING) : BOOLEAN;
(* Liest den Code in codeTextFile, entschlsselt ihn ber den geheimen *)
(* Vektor vector und schreibt den Klartext in klarTextFile.            *)
(* Liefert TRUE, wenn alles geklappt hat, sonst FALSE.                 *)
CONST
  WordLength = KeyLength DIV 8; (* Wortlnge bei 8 Bits pro Zeichen *)
VAR
  count              : INTEGER;
  code               : CodeT;
  klar               : STRING;
  klarFile           : FILE OF CHAR;
  codeFile           : FILE OF CodeT;

  PROCEDURE Decodieren(VAR klar : STRING; code : CodeT);
  (* Decodiert das "Wort" in code und legt das Ergebnis in klar ab *)
  TYPE
    ByteT = ARRAY[1..8] OF BOOLEAN;    (* Zhler hier mal von 1 bis 8 *)
    BitFolgeT = ARRAY[1..KeyLength] OF BOOLEAN;
  VAR
    bitfolge : BitFolgeT;

    PROCEDURE Code2Bitfolge(code : CodeT; VAR bitfolge : BitFolgeT);
    VAR
      bitCount : INTEGER;
    BEGIN
      FOR bitCount:= KeyLength DOWNTO 1 DO
      BEGIN
        IF (vector[bitCount] <= code) THEN
        BEGIN
          DEC(code,vector[bitCount]);
          bitfolge[bitCount]:= TRUE;
        END
        ELSE
          bitfolge[bitCount]:= FALSE
      END;
    END;

    PROCEDURE Bitfolge2String(bitfolge : BitFolgeT; VAR str : STRING);
    (* Konvertiert die Bitfolge in eine Zeichenfolge *)
    VAR
      zeichenCount, count, position : INTEGER;
      byte                       : ByteT;

      PROCEDURE Byte2Char(byte : ByteT; VAR zeichen : CHAR);
      (* Macht aus einer Bitfolge von 8 Bits ein Zeichen *)
      VAR
        ord_zahl, count, calc : INTEGER;
      BEGIN
        (* Zeichenwert berechenen *)
        ord_zahl:= 0;
        FOR count:= 1 TO 8 DO
        BEGIN
          IF byte[count] THEN
          BEGIN
            calc:= ROUND( Power(2.0,(count-1)) );
            INC(ord_zahl,calc);
          END
        END;
        zeichen:= CHR(ord_zahl);
      END;

    BEGIN
      str:= '';
      (* Fr alle Zeichen in der zu findenden Zeichenfolge *)
      FOR zeichenCount:= 1 TO WordLength DO
      BEGIN
        (* Bestimme die Position des Byte dieses Zeichens in der Bitfolge *)
        position:= zeichenCount * 8 - 8;
        (* Nimm das "Byte" aus der Bitfolge *)
        FOR count:= 1 TO 8 DO
          byte[count]:= bitfolge[position+count];
        (* Wandle das Byte in ein Zeichen um *)
        Byte2Char(byte,str[zeichenCount]);
        Write(str[zeichenCount]);
      END;
    END;

  BEGIN
    (* Code zurckrechnen *)
    code:= ( code * einheit ) MOD divisor;
    (* Mache Bitfolge aus dem Code *)
    Code2Bitfolge(code,bitfolge);
    (* Mache "Wort" aus der Bitfolge *)
    Bitfolge2String(bitfolge,klar);
  END;

BEGIN
  (* Prfen, ob Codetext-Datei existiert *)
  IF NOT Exists(codeTextFile) THEN
  BEGIN
    (* Negative Rckmeldung *)
    Entschluesseln:= FALSE;
    Exit;
  END;
  (* Dateien aufmachen *)
  Assign(klarFile,klarTextFile);
  Rewrite(klarFile);
  Assign(codeFile,codeTextFile);
  Reset(codeFile);
  (* Decodierungsschleife *)
  (* FIO.EOF:= FALSE; *)
  WHILE NOT EOF(codeFile) DO
  BEGIN
    (* Codezahl lesen (code) *)
    Read(codeFile,code);
    Write('Code ',code,' --> ');
    (* Decodieren *)
    Decodieren(klar,code);
    WriteLn(klar);
    (* Klartext rausschreiben *)
    FOR count:= 1 TO WordLength DO
      Write(klarFile,klar[count]);
  END;
  (* Dateien schlieen *)
  Close(klarFile);
  Close(codeFile);
  (* Positive Rckmeldung *)
  Entschluesseln:= TRUE;
END;

END.
